import curses, os, sys, types, signal
import kaa.tools
import kaa.log
from kaa import options, version, consts, config
from . import app, keydef, frame

from .. import document
from kaa import fileio

CURSES_MOUSEINTERVAL = 200
CURSES_ESCDELAY = '50'

def _init(stdscr):
    if not hasattr(stdscr, 'get_wch'):
        raise RuntimeError(
                'Kaa requires curses library with wide charater support.')
        
    curses.start_color()
    curses.use_default_colors()

    curses.raw()
    curses.nonl()
    stdscr.leaveok(1)
#    curses.mousemask(curses.ALL_MOUSE_EVENTS)
#    curses.mouseinterval(CURSES_MOUSEINTERVAL)


def _restore():
    curses.noraw()
    curses.nl()


def run_userinit(fname):
    if os.path.isfile(fname):
        with open(fname) as f:
            src = f.read()

        code = compile(src, fname, 'exec')
        module = types.ModuleType('__kaa__')
        exec(code, module.__dict__)

        sys.modules['__kaa__'] = module


def main(stdscr):
    conf = config.Config(opt)
    sys.path.insert(0, conf.KAADIR)

    kaa.log.init(conf.LOGDIR)

    _init(stdscr)
    try:
        keydef.init(conf)

        kaa.app = app.CuiApp(conf)
        kaa.app.storage = fileio.FileStorage()
        
        if not opt.no_init:
            fname = opt.init_script
            if not fname:
                fname = os.path.join(kaa.app.config.KAADIR, '__kaa__.py')
            run_userinit(fname)

        mainframe = frame.MainFrame(stdscr)
        kaa.app.init(mainframe)

        if not opt.file:
            # no file args. show new document.
            doc = fileio.newfile(provisional=True)
            kaa.app.show_doc(doc)
        else:
            dirname = None
            for filename in opt.file:
                if os.path.isdir(filename):
                    if not dirname:
                        dirname = filename
                else:
                    doc = kaa.app.storage.openfile(filename)
                    kaa.app.show_doc(doc)
            
            if dirname:
                from kaa.ui.selectfile import selectfile
                def cb(filename, encoding, newline):
                    if filename:
                        doc = kaa.app.storage.openfile(
                                filename, encoding, newline)
                        kaa.app.show_doc(doc)
                    else:
                        if not kaa.app.mainframe.childframes:
                            doc = fileio.newfile(provisional=True)
                            kaa.app.show_doc(doc)
                    
                selectfile.show_fileopen(dirname, cb)
                
        kaa.app.run()
        kaa.app.on_shutdown()

        mainframe.destroy()

    finally:
        _restore()

def handle_term(signum, frame):
    sys.exit(signum+0x80)

COLOR_ENVS = ('COLORTERM', 'XTERM_VERSION', 'ROXTERM_ID',
             'KONSOLE_DBUS_SESSION')

def _init_term():
    if opt.term:
        os.environ['TERM'] = opt.term

    # http://fedoraproject.org/wiki/Features/256_Color_Terminals
    for env in COLOR_ENVS:
        if os.environ.get(env):
            has_color = True
            break
    else:
        has_color = False
        
    if has_color:
        term = os.environ.get('TERM')
            
        if term in ('xterm', 'screen', 'Eterm'):
            term = term+'-256color'
            os.environ['TERM'] = term
    
        if os.environ.get('TERM', '') == 'screen-256color':
            termcap = os.environ.get('TERMCAP')
            if termcap:
                os.environ['TERMCAP'] = termcap.replace('Co#8', 'Co#256')
    
def run():
    if sys.version_info[:2] < (3, 3):
        raise RuntimeError('kaa requires Python 3.3 or later')
    
    if not getattr(__builtins__, 'kaa_freeze', False):
        try:
            setproctitle = __import__('setproctitle')
            setproctitle.setproctitle('kaa')
        except:
            pass

    signal.signal(signal.SIGTERM, handle_term)
    
    parser = options.build_parser()

    global opt
    opt = parser.parse_args()

    if opt.show_version:
        print(version.version_info())
        return

    if not os.environ.get('ESCDELAY'):
        os.environ['ESCDELAY'] = CURSES_ESCDELAY

    _init_term()
    curses.wrapper(main)


if __name__ == '__main__':
    run()



import os
import kaa
from kaa import document
from kaa.filetype.default import defaultmode
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.command import command, norec, norerun
from kaa.keyboard import *
from kaa.commands import (appcommand, filecommand)
from kaa.ui.selectlist import selectlist
import kaa.ui.pythondebug.port
from kaa.ui.inputline import inputlinemode


def show_expr(port, depth):
    def callback(w, s):
        s = s.strip()
        if s:
            kaa.app.config.hist_pythondebug_expr.add(s)
            port.show_expr(depth, s)

    hist = [s for s, info in kaa.app.config.hist_pythondebug_expr.get()]
    doc = inputlinemode.InputlineMode.build('Python expression:', 
                callback, history=hist)
    kaa.app.show_dialog(doc)
    

breakpoints_keys = {
    'd': 'breakpoints.delete',
    (ctrl, 'h'): 'breakpoints.delete',
    delete: 'breakpoints.delete',
    backspace: 'breakpoints.delete',
}

class BreakPoints(selectlist.SelectItemList):
    SEP = '\n'
    caption = 'Break points'

    @classmethod
    def build(cls, port):
        doc = super().build()
        doc.mode.show_breakpoints(port)
        kaa.app.messagebar.set_message("Hit 'd' key to remove breakpoint")
        return doc

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(breakpoints_keys)

    def show_breakpoints(self, port):
        self.port = port
        breakpoints = self.port.get_breakpoints()
        breakpoints.sort(key=lambda o:(o.filename, o.lineno))
        items = []
        for bp in breakpoints:
           caption = '{bp.filename}:{bp.lineno}'.format(bp=bp)
           c = selectlist.SelectItem(
                'selectitem', 'selectitem-active', caption, bp)
           items.append(c)
            
        self.update_doc(items)

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    @command('breakpoints.delete')
    @norec
    @norerun
    def delete_bp(self, wnd):
        if not self.cursel:
            return
        bp = self.cursel.value
        for doc in document.Document.all:
            if doc.get_filename() == bp.filename:
                if bp in doc.marks:
                    del doc.marks[bp]
                break

        index = self.items.index(self.cursel)
        items = [item for item in self.items if item is not self.cursel]
        self.update_doc(items)
        if not self.items:
            sel = None
        elif index < len(self.items):
            sel = self.items[index]
        else:
            sel = self.items[-1]
        
        self.update_sel(wnd, sel)

DebugThemes = {
    'basic':
        Theme([
            Style('line', 'Cyan', None),
            Style('line-active', 'Base02', 'Yellow'),
            Style('filename', 'Default', None),
            Style('lineno', 'Default', None),
            Style('funcname', 'Default', None),
            Style('dirname', 'Default', None),
        ]),
}


callstack_keys = {
    down: 'callstack_keys.next',
    (ctrl, 'n'): 'callstack_keys.next',
    (ctrl, 'f'): 'callstack_keys.next',
    tab: 'callstack_keys.next',
    up: 'callstack_keys.prev',
    (ctrl, 'p'): 'callstack_keys.prev',
    (ctrl, 'b'): 'callstack_keys.prev',
    (shift, tab): 'callstack_keys.prev',
}


class PythonCallStackMode(dialogmode.DialogMode):
    SHOW_CURSOR = False
    autoshrink = True
    cursel = 0

    def setup(self):
        super().setup()
        self.updated_wnds = set()

    def init_themes(self):
        super().init_themes()
        self.themes.append(DebugThemes)

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(callstack_keys)
    
    def init_commands(self):
        super().init_commands()

        self.app_commands = appcommand.ApplicationCommands()
        self.register_command(self.app_commands)

        self.file_commands = filecommand.FileCommands()
        self.register_command(self.file_commands)

    def close(self):
        self._restore_highlight()
        super().close()
        self.port = None
        
    def on_str(self, wnd, s):
        pass

    def calc_position(self, wnd):
        height = wnd.screen.get_total_height()
        height = min(height, 
            (wnd.mainframe.height - wnd.mainframe.MESSAGEBAR_HEIGHT)//2)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top+height

    def build(self, stack):
        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())

        self.stack = tuple(stack or ())
        with dialogmode.FormBuilder(self.document) as f:
            f.append_text('button', '[&Step]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_step)
    
            f.append_text('button', '[&Next]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_next)
    
            f.append_text('button', '[&Return]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_return)
    
            f.append_text('button', '[&Continue]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_continue)
    
            f.append_text('button', '[&Expr]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_expr)
    
            f.append_text('button', '[&Breakpoint]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_breakpoint)
    
            f.append_text('button', '[&Quit]', 
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_quit)
    
            f.append_text('default', '\n')
    
            for n, (fname, lno, funcname, lines) in enumerate(self.stack):
                s = self.document.endpos()
                
                line = (lines[0] if lines else '').strip()
                if not line:
                    line = '(empty line)'
                f.append_text('line', line.replace('&', '&&')+'\n')
                
                t = self.document.endpos()
                self.document.marks[('stack', n)] = (s, t)
                
                dirname, filename = os.path.split(fname)
                f.append_text('filename', filename.replace('&', '&&'))
                f.append_text('default', ':')
    
                f.append_text('lineno', str(lno))
                f.append_text('default', ':')
    
                f.append_text('funcname', funcname.replace('&', '&&'))
                f.append_text('default', ':')
    
                f.append_text('dirname', dirname.replace('&', '&&')+'\n')


    def update(self, wnd, stack):
        self.build(stack)
        self.update_sel(wnd, 0)
        wnd.mainframe.on_console_resized()

    def update_sel(self, wnd, n):
        if self.cursel != n:
            f, t = self.document.marks.get(
                    ('stack', self.cursel), (None, None))
            if f is not None:
                self.document.styles.setints(f, t, self.get_styleid('line'))
                
        f, t = self.document.marks.get(('stack', n), (None, None))
        if f is not None:
            self.document.styles.setints(
                f, t, self.get_styleid('line-active'))
            
        self.cursel = n
        if n < len(self.stack):
            (fname, lno, funcname, lines) = self.stack[n]
            self.show_line(fname, lno)
            wnd.activate()

            if f is not None:
                wnd.screen.locate(f, middle=True)
            
    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    def _restore_highlight(self):
        for w in self.updated_wnds:
            w.clear_line_overlay()
        self.updated_wnds = set()
        
    def _locate_doc(self, wnd, doc, lineno):
        pos = doc.get_lineno_pos(lineno)
        tol = doc.gettol(pos)

        wnd.clear_line_overlay()
        wnd.set_line_overlay(tol, 'current_row')

        wnd.screen.locate(tol, middle=True, align_always=True)
        wnd.cursor.setpos(tol, middle=True)
        wnd.activate()
        self.updated_wnds.add(wnd)

    def show_line(self, filename, lineno):
        filename = os.path.abspath(filename)
        doc = document.Document.find_filename(filename)
        if doc:
            wnd = doc.wnds[0]
            wnd.activate()
        else:
            doc = kaa.app.storage.openfile(filename)
            wnd = kaa.app.show_doc(doc)
        
        self._locate_doc(wnd, doc, lineno)
       
    @command('callstack_keys.prev')
    @norec
    @norerun
    def stack_prev(self, wnd):
        if self.cursel:
            self.update_sel(wnd, self.cursel-1)
            
    @command('callstack_keys.next')
    @norec
    @norerun
    def stack_next(self, wnd):
        if self.cursel < len(self.stack)-1:
            self.update_sel(wnd, self.cursel+1)

    def on_step(self, wnd):
        self.port.set_step()

    def on_next(self, wnd):
        self.port.set_next()

    def on_return(self, wnd):
        self.port.set_return()

    def on_continue(self, wnd):
        self.port.set_continue()

    def on_expr(self, wnd):
        show_expr(self.port, self.cursel)

    def on_breakpoint(self, wnd):
        doc = BreakPoints.build(self.port)
        kaa.app.show_dialog(doc)
        
    def on_quit(self, wnd):
        self.port.close()
        
def show_callstack(port, stack):
    # update current breakpoints
    kaa.ui.pythondebug.port.update_breakpoints()

    buf = document.Buffer()
    doc = document.Document(buf)
    doc.set_title('Python call stack')
    mode = PythonCallStackMode()
    mode.port = port

    doc.setmode(mode)
    mode.build(stack)
    
    dlg = kaa.app.show_inputline(doc)
    ret = dlg.get_label('editor')
    mode.update_sel(ret, 0)
    return ret



import os, unicodedata, sys, locale
import kaa
import kaa.log
from kaa import document, encodingdef
from kaa.filetype.default import defaultmode
from kaa import consts

class FileStorage:
    def adjust_encoding(self, encoding):
        if encoding.lower() == 'shiftjis':
            return 'cp932'
        return encoding
        
    def guess_japanese_encoding(self, filename):
        import pyjf3
        ret = pyjf3.guess(open(filename, 'rb').read())
        
        if ret == pyjf3.ASCII:
            return kaa.app.config.DEFAULT_ENCODING
        elif ret == pyjf3.UTF8:
            return 'utf-8'
        elif ret == pyjf3.SJIS:
            return 'cp932'
        elif ret == pyjf3.EUC:
            return 'euc-jp'
        elif ret == pyjf3.JIS:
            return 'iso-2022-jp'
        elif ret == pyjf3.UTF16_LE:
            return 'utf-16le'
        elif ret == pyjf3.UTF16_BE:
            return 'utf-16be'
        
    def get_textio(self, fileinfo, filemustexists=False):
        try:
            # use surrogateescape to preserve file contents intact.
            textio = open(fileinfo.fullpathname, 'r',
                        encoding=self.adjust_encoding(fileinfo.encoding), 
                        errors='surrogateescape', 
                        newline=fileinfo.nlchars)
        
        except FileNotFoundError:
            if filemustexists:
                raise
            return None

        return textio

    def get_fileinfo(self, filename, encoding=None, newline=None):
        if sys.platform == 'darwin':
            filename = unicodedata.normalize('NFC', filename)
    
        if encoding is None:
            encoding = kaa.app.config.DEFAULT_ENCODING
    
        if newline is None:
            newline = kaa.app.config.DEFAULT_NEWLINE
    
        fileinfo = FileInfo()

        fullpath = os.path.abspath(filename)
        dirname, filename = os.path.split(filename)

        fileinfo.storage = self
        fileinfo.fullpathname = fullpath
        fileinfo.dirname = dirname
        fileinfo.filename = filename

        try:
            fileinfo.stat = os.stat(fullpath)
        except FileNotFoundError:
            fileinfo.stat = None

        fileinfo.encoding = encoding
        fileinfo.newline = newline
        
        return fileinfo
        
    def listdir(self, dirname):
        dirs = []
        files = []
        for name in os.listdir(dirname):
            if os.path.isdir(os.path.join(dirname, name)):
                dirs.append(name)
            else:
                files.append(name)
        if sys.platform == 'darwin':
            dirs = [unicodedata.normalize('NFC', n) for n in dirs]
            files = [unicodedata.normalize('NFC', n) for n in files]
        return dirs, files

    def _save_hist(self, filename):
        abspath = os.path.abspath(filename)
        dirname = os.path.dirname(abspath)
        if dirname.endswith(os.pathsep):
            dirname = dirname[:-1]

        kaa.app.config.hist_files.add(abspath)
        kaa.app.config.hist_dirs.add(dirname)
        kaa.app.last_dir = dirname

    def openfile(self, filename, encoding=None, newline=None, nohist=False,
                    filemustexists=False):
        if not nohist:
            self._save_hist(filename)
        return openfile(filename, encoding, newline, filemustexists)

    def save_document(self, doc, filename, encoding=None, newline=None, 
                        nohist=False):
        # todo: move most of codes to mode class.
        if not nohist:
            self._save_hist(filename)

        if not doc.fileinfo:
            fileinfo = kaa.app.storage.get_fileinfo(filename, encoding, newline)
        else:
            if encoding is not None:
                doc.fileinfo.encoding = encoding

            if doc.fileinfo.encoding is None:
                doc.fileinfo.encoding = kaa.app.config.DEFAULT_ENCODING

            if newline is not None:
                doc.fileinfo.newline = newline
                
            if doc.fileinfo.newline is None:
                doc.fileinfo.newline = kaa.app.config.DEFAULT_NEWLINE

        with open(filename, 'w', 
                 encoding=self.adjust_encoding(doc.fileinfo.encoding),
                 newline=consts.NEWLINE_CHARS[doc.fileinfo.newline],
                 errors='surrogateescape') as f:

            # TODO: save as another file and rename.
            # TODO: fsync
            f.write(doc.gettext(0, doc.endpos()))

        fileinfo = self.get_fileinfo(filename, doc.fileinfo.encoding, 
                                            doc.fileinfo.newline)
        doc.mode.on_file_saved(fileinfo)
        
        doc.fileinfo = fileinfo
        doc.set_title(None)
        doc.provisional = False
        if doc.undo:
            doc.undo.saved()
        
        mode = select_mode(filename)
        if type(doc.mode) is not mode:
            doc.setmode(mode())

        dir, file = os.path.split(filename)
        if not dir.endswith(os.path.sep):
            dir += os.path.sep

        kaa.app.messagebar.set_message('Written to {}({})'.format(file, dir))


class FileInfo:
    storage = None
    fullpathname = ''
    dirname = ''
    filename = ''
    stat = None
    encoding = None
    newline = None

    @property
    def nlchars(self):
        return consts.NEWLINE_CHARS[self.newline]

    def __init__(self):
        self.encoding = kaa.app.config.DEFAULT_ENCODING
        self.newline= kaa.app.config.DEFAULT_NEWLINE

    def save_screenloc(self, pos):
        if self.fullpathname:
            info = kaa.app.config.hist_filedisp.find(self.fullpathname)
            if not info:
                info = {}
            info['pos'] = pos
            kaa.app.config.hist_filedisp.add(
                self.fullpathname, info)

    def check_update(self):
        if self.fullpathname:
            try:
                stat = os.stat(self.fullpathname)
            except Exception:
                return
    
            if stat.st_mtime == self.stat.st_mtime:
                return
            
            if stat.st_size == self.stat.st_size:
                return
            
            self.stat = stat
            return True

def select_mode(filename):
    ext = os.path.splitext(filename)[1].lower()

    for pkg in kaa.app.config.get_mode_packages():
        if ext in getattr(pkg, 'FILE_EXT', ()):
            return pkg.get_modetype()

    return defaultmode.DefaultMode


def openfile(filename, encoding=None, newline=None, filemustexists=False):
    # Open file
    fileinfo = kaa.app.storage.get_fileinfo(filename, encoding, newline)
    textio = kaa.app.storage.get_textio(fileinfo, filemustexists)

    buf = document.Buffer()
    if textio:
        src = textio.read()
        if fileinfo.nlchars is not None:
            src = src.replace(fileinfo.nlchars, '\n')
        buf.insert(0, src)

    doc = document.Document(buf)
    doc.fileinfo = fileinfo
    doc.setmode(select_mode(filename)())

    dir, file = os.path.split(fileinfo.fullpathname)
    if not dir.endswith(os.path.sep):
        dir += os.path.sep

    kaa.app.messagebar.set_message('Read from {}({})'.format(file, dir))

    return doc

NUM_NEWFILE = 1
def newfile(mode=None, s='', provisional=False):
    buf = document.Buffer()
    if s:
        buf.insert(0, s)

    doc = document.Document(buf)
    doc.provisional = provisional

    global NUM_NEWFILE
    doc.title = '<Untitled {}>'.format(NUM_NEWFILE)
    NUM_NEWFILE += 1

    doc.fileinfo = FileInfo()
    if not mode:
        mode = defaultmode.DefaultMode()
    doc.setmode(mode)
    return doc



import logging, logging.handlers, os
import kaa

def init(logdir):
    logfile = os.path.join(logdir, 'kaa.log')

    logging.basicConfig(
            format="%(asctime)s %(name)s:%(levelname)s:%(message)s",
            datefmt='%Y/%m/%d %H:%M:%S',
            level=logging.DEBUG,
            handlers=[logging.handlers.RotatingFileHandler(
                filename=logfile, encoding='utf-8',
                maxBytes=10*1024*1024)])

    global LOG, debug, info, warning, error, critical, exception
    LOG = logging.getLogger('kaa')


    debug = LOG.debug
    info = LOG.info
    warning = LOG.warning
    error = LOG.error
    critical = LOG.critical
    exception = LOG.exception




MENUS = {
    'MAIN':
        [['[&File]', 'FILE', None],
         ['[&Edit]', 'EDIT', None],
         ['[&Code]', 'CODE', None],
         ['[&Macro]', 'MACRO', None],
         ['[&Tools]', 'TOOLS', None],
         ['[&Window]', 'WINDOW', None]],

    'FILE':
        [['&New', None, 'file.new'],
         ['&Open', None, 'file.open'],
         ['File &Info', None, 'file.info'],
         ['View &Diff', None, 'file.viewdiff'],
         ['&Save', None, 'file.save'],
         ['Save &As', None, 'file.saveas'],
         ['&Close', None, 'file.close'],
         ['Save a&ll', None, 'file.save.all'],
         ['Clos&e all', None, 'file.close.all'],
         ['[&Recently]', 'RECENTLY-USED-FILES', None],
         ['&Quit', None, 'file.quit']],
         
    'RECENTLY-USED-FILES':
        [['Recently used &Files', None, 'file.recently-used-files'],
         ['Recently used &Dirs', None, 'file.recently-used-directories']],

    'EDIT':
        [['&Cut', None, 'edit.cut'],
         ['C&opy', None, 'edit.copy'],
         ['&Paste', None, 'edit.paste'],
         ['Paste &History', None, 'edit.clipboard-history'],
         ['&Undo', None, 'edit.undo'],
         ['&Redo', None, 'edit.redo'],
         ['&Search', None, 'search.showsearch'],
         ['R&eplace', None, 'search.showreplace'],
         ['Co&mplete', None, 'edit.word-complete'],
         ['[&Convert]', 'EDIT-CONVERT', None]],

    'EDIT-CONVERT':
        [['&Upper', None, 'edit.conv.upper'],
         ['&Lower', None, 'edit.conv.lower'],
         ['&Normalization', None, 'edit.conv.nfkc'],
         ['&Full-width', None, 'edit.conv.full-width']],

    'MACRO':
        [['&Start Record', None, 'macro.start-record'],
         ['&End Record', None, 'macro.end-record'],
         ['&Run Macro', None, 'macro.run']],
        
    'TOOLS':
        [['&Grep', None, 'search.showgrep'],
         ['Paste &Lines', None, 'edit.paste-lines'],
         ['&Shell command', None, 'tools.execute-shell-command'],
         ['&Make', None, 'tools.make'],
         ['&Python console', None, 'python.console'],
         ['Python &Debugger', None, 'python.debugger.run'],
         ['Python debugger s&erver', None, 'python.debugger.server'],
         ],
         
    'WINDOW':
        [['&Frame list', None, 'app.show-framelist'],
         ['Split &vert', None, 'editor.splitvert'],
         ['Split &horz', None, 'editor.splithorz'],
         ['&Move separator', None, 'editor.moveseparator'],
         ['&Next window', None, 'editor.nextwindow'],
         ['&Prev window', None, 'editor.prevwindow'],		
         ['&Join window', None, 'editor.joinwindow'],
         ['[&Switch file]', 'CHANGE-WINDOW', None]],

     'CHANGE-WINDOW':
         [['&Switch file', None, 'editor.switchfile'],
          ['&New file here', None, 'file.new-to'],
          ['&Open file here', None, 'file.open-to'],
          ['Recently used &Files', None, 'file.recently-used-files-to'],
          ['Recently used &Dirs', None, 'file.recently-used-directories-to']],
}



import os, importlib, sqlite3, json
import kaa
from kaa import consts, utils

class KaaHistoryStorage:
    def __init__(self, filename):
        self.filename = filename
        self.conn = sqlite3.connect(filename)
        
    def close(self):
        self.conn.close()
        
    def add_history(self, history):
        history.set_storage(self)

    

class History:
    MAX_HISTORY = 1000

    def __init__(self, name, storage):
        self.name = name
        self.table_name = 'hist_'+self.name
        self.storage = None

        storage.add_history(self)
        
    @utils.ignore_errors
    def close(self):
        ret = self.storage.conn.execute('''
            SELECT id FROM {} ORDER BY id DESC LIMIT ?
            '''.format(self.table_name), (self.MAX_HISTORY, ))
        recs =[value for value in ret]
        if recs:
            last, = recs[-1]
            self.storage.conn.execute('''
                DELETE FROM {} WHERE id < ?
                '''.format(self.table_name), (last,))
            self.storage.conn.commit()
            
    def set_storage(self, storage):
        storage.conn.execute('''
            CREATE TABLE IF NOT EXISTS {}
                (id INTEGER PRIMARY KEY,
                value TEXT,
                info  TEXT)'''.format(self.table_name))
        self.storage = storage

    @utils.ignore_errors
    def add(self, value, info=None):
        if not value:
            return

        ret = self.storage.conn.execute('''
            DELETE FROM {} WHERE value = ?'''.format(self.table_name),
            (value, ))

        self.storage.conn.execute('''
            INSERT INTO {}(value, info) VALUES(?, ?)
            '''.format(self.table_name), (value, json.dumps(info)))

        self.storage.conn.commit()

    def get(self):
        ret = self.storage.conn.execute('''
            SELECT value, info FROM {} ORDER BY id DESC LIMIT ?
            '''.format(self.table_name),
            (self.MAX_HISTORY,))
        return [(value, json.loads(info)) for value,info in ret]

    @utils.ignore_errors
    def find(self, value):
        ret = self.storage.conn.execute('''
            SELECT info FROM {} WHERE value=?
            '''.format(self.table_name),
            (value,))

        for info, in ret:
            return json.loads(info)
        
class Config:
    FILETYPES = [
        'kaa.filetype.default',
        'kaa.filetype.python',
        'kaa.filetype.html',
        'kaa.filetype.javascript',
        'kaa.filetype.css',
        'kaa.filetype.diff',
        'kaa.filetype.rst',
        'kaa.filetype.markdown',
        'kaa.filetype.c',
    ]

    DEFAULT_NEWLINE = 'auto'
    DEFAULT_ENCODING = 'utf-8'
    AMBIGUOUS_WIDTH = 1
    
    def __init__(self, option):
        kaadir = os.path.abspath(
                 os.path.expandvars(
                 os.path.expanduser(consts.KAA_DIR)))

        if not os.path.exists(kaadir):
            os.makedirs(kaadir)

        logdir = os.path.join(kaadir, consts.LOGDIR)
        if not os.path.exists(logdir):
            os.mkdir(logdir)

        histdir = os.path.join(kaadir, consts.HISTDIR)
        if not os.path.exists(histdir):
            os.mkdir(histdir)

        self.KAADIR = kaadir
        self.LOGDIR = logdir
        self.HISTDIR = histdir

        self.palette = option.palette
        
    def init_history(self):
        self.hist_storage = KaaHistoryStorage(
            os.path.join(self.HISTDIR, consts.HIST_DBNAME))

        self.hist_files = History('filename', self.hist_storage)
        self.hist_filedisp = History('filedisp', self.hist_storage)
        self.hist_dirs = History('dirname', self.hist_storage)
        self.hist_searchstr = History('search_text', self.hist_storage)
        self.hist_replstr = History('repl_text', self.hist_storage)
        self.hist_grepstr = History('grep_text', self.hist_storage)
        self.hist_grepdir = History('grep_dir', self.hist_storage)
        self.hist_grepfiles = History('grep_filename', self.hist_storage)
        self.hist_shellcommands = History('shellcommands', self.hist_storage)
        self.hist_makecommands = History('makecommands', self.hist_storage)
        self.hist_pythondebug_expr = History('pythondebug_expr', self.hist_storage)
        self.hist_pythondebugcommands = History('pythondebug_cmdline', self.hist_storage)

    def close(self):
        self.hist_files.close()
        self.hist_dirs.close()
        self.hist_searchstr.close()
        self.hist_replstr.close()
        self.hist_grepstr.close()
        self.hist_grepdir.close()
        self.hist_grepfiles.close()
        self.hist_shellcommands.close()
        self.hist_makecommands.close()
        self.hist_pythondebug_expr.close()
        self.hist_pythondebugcommands.close()
        
    def get_mode_packages(self):
        for pkgname in self.FILETYPES:
            try:
                pkg = importlib.import_module(pkgname)
            except Exception:
                kaa.log.exception('Error loading filetype: '+repr(pkgname))
            else:
                yield pkg
   


import copy
from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import Tokenizer, Token, Span, Keywords, EndSection, SingleToken
from kaa.theme import Theme, Style
from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *

MarkdownThemes = {
    'basic':
        Theme([
            Style('header', 'Blue', None),
            Style('hr', 'Green', None),
            Style('strong', 'Magenta', None),
            Style('emphasis', 'Blue', None),
            Style('literal', 'Cyan', None),
            Style('reference', 'Red', None),
            Style('role', 'Cyan', None),
            Style('substitution', 'Green', None),
        ]),
}

class LinkToken(Token):
    RE_END = gre.compile('^\S', gre.M+gre.X)
    
    def re_start(self):
        return r'\['

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.text = self.assign_tokenid(tokenizer, self.stylename)
        self.url = self.assign_tokenid(tokenizer, self.stylename)
        self.link = self.assign_tokenid(tokenizer, self.stylename)

    def get_close(self, doc, pos, chars):
        r = gre.compile(r'\\.|'+'|'.join('\\'+c for c in chars))
        while True:
            m = r.search(doc.buf, pos)
            if not m:
                return '', doc.endpos()

            h = m.group()
            if h[0] == '\\':
                pos = m.end()
            else:
                return h, m.end()
        
    def on_start(self, tokenizer, doc, pos, match):
        f = match.start()
        c, t = self.get_close(doc, match.end(), ']')
        if t != doc.endpos():
            next = doc.gettext(t, t+1)
            if next == ':':
                yield (f, t+1, self.text)
                return t+1, None, False
            elif next == ' ':
                t += 1
        yield (f, t, self.text)
    
        m = gre.compile(r'\s*(\[|\()').match(doc.buf, t)
        if not m:
            return t, None, False

        c = m.group(1)
        if c == '[':
            c, end = self.get_close(doc, m.end(), ']')
            yield t, end, self.link
        else:
            c, end = self.get_close(doc, m.end(), '")')
            if c == '"':
                c, end = self.get_close(doc, end, '"')
                c, end = self.get_close(doc, end, ')')
            yield t, end, self.url

        return end, None, False

class ImageToken(LinkToken):
    def re_start(self):
        return r'!\['

class MDInline(Span):
    WS = ' \t\r\n'
    
    def on_start(self, tokenizer, doc, pos, match):
        if ((pos >= doc.endpos()-1) or
            (doc.gettext(pos+1, pos+2) in self.WS)):

            yield pos, pos+1, tokenizer.nulltoken
            return pos+1, None, False
                
        ret = yield from super().on_start(tokenizer, doc, pos, match)
        return ret

    
HEADERS = r'=-'
def build_tokenizer():
    MARKDOWNTOKENS = namedtuple('markdowntokens', 
                            ['escape', 'header1', 'header2',
                             'hr', 'link', 'image', 
                             'strong1', 'strong2', 
                             'emphasis1', 'emphasis2', 'code1', 'code2'])

    return Tokenizer(MARKDOWNTOKENS(
            # escape
            SingleToken('md-escape', 'default', [r'\\.']),

            # header
            SingleToken('md-header1', 'header', 
                    [r'^.+\n(?P<H1>[{}])(?P=H1)+$'.format(HEADERS)]),
            SingleToken('md-header2', 'header', [r'^\#{1,6}.*$']),

            # hr
            SingleToken('md-hr', 'hr', [r'^(\-{3,}|_{3,}|\*{3,})$']),
            
            # link
            LinkToken('md-link', 'reference'),
            
            # image
            ImageToken('md-image', 'reference'),
            
            # strong
            MDInline('md-strong1', 'strong', r'\*\*', r'\*\*|$', escape='\\'),
            MDInline('md-strong2', 'strong', r'__', r'__|$', escape='\\'),

            # emphasis
            MDInline('md-emphasis1', 'emphasis', r'\*', r'\*|$', escape='\\'),
            MDInline('md-emphasis2', 'emphasis', r'_', r'_|$', escape='\\'),

            # code
            MDInline('md-code1', 'literal', r'``', r'``', escape='\\'),
            MDInline('md-code2', 'literal', r'`', r'`', escape='\\'),
        ))



MDMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

md_keys = {
    (ctrl, 't'): 'toc.showlist',
}

class MarkdownMode(defaultmode.DefaultMode):
    MODENAME = 'Markdown'

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [md_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(MDMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(MarkdownThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]

    HEADER1 = r'^(?P<TITLE>.+)\n(?P<H1>[{}])(?P=H1)+$'.format(HEADERS)
    HEADER2 = r'^(?P<H2>\#{1,6})(?P<TITLE2>.+)$'
    
    RE_HEADER = gre.compile('|'.join([HEADER1, HEADER2]), gre.X|gre.M)

    def get_headers(self):
        stack = []
        pos = 0
        while True:
            m = self.RE_HEADER.search(self.document.buf, pos)
            if not m:
                break
    
            pos = m.end()
            name = m.group('TITLE') or m.group('TITLE2')
            name = name.strip()

            bar = m.group('H1')
            if bar:
                if bar == '=':
                    level = 0
                else:
                    level = 1
            else:
                level = len(m.group('H2'))-1

            if not stack:
                header = self.HeaderInfo('namespace', None, name, name, 
                            None, m.start())
                yield header
                stack.append((level, header))
                continue

            for i, (parent_level, info) in enumerate(stack):
                if level <= parent_level:
                    del stack[i:]
                    break

            if stack:
                parent = stack[-1][1]
            else:
                parent = None

            header = self.HeaderInfo(
                'namespace', parent, name, 
                name, None, m.start())
            yield header
            stack.append((level, header))
            

    @command('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        headers = list(self.get_headers())
        toclistmode.show_toclist(wnd, headers)




import itertools, unicodedata, collections

import gappedbuf.re
import kaa
from kaa import keyboard, editmode
import kaa.log
from kaa import highlight
from kaa import theme
from kaa import screen

class SearchOption:
    RE = gappedbuf.re
    def __init__(self):
        self.text = ''
        self.replace_to = ''
        self.ignorecase = True
        self.word = False
        self.regex = False

    def clone(self):
        ret = self.__class__()
        ret.text = self.text
        ret.replace_to = self.replace_to
        ret.ignorecase = self.ignorecase
        ret.word = self.word
        ret.regex = self.regex
        return ret
        
    def get_regex(self):
        text = self.text
        if not self.regex:
            text = self.RE.escape(text)
        if self.word:
            text = r'\b'+text+r'\b'

        opt = self.RE.MULTILINE
        if self.ignorecase:
            opt += self.RE.IGNORECASE

        regex = self.RE.compile(text, opt)
        return regex

SearchOption.LAST_SEARCH = SearchOption()

class ModeBase:
    CLOSE_ON_DEL_WINDOW = True

    SCREEN_NOWRAP = False
    NO_WRAPINDENT = False
    SCREEN_BUILD_ENTIRE_ROWS = False
    SHOW_LINENO = False
    USE_UNDO = True
    DOCUMENT_MODE = False
    HIGHLIGHT_CURSOR_ROW = False
    SHOW_CURSOR = True
    
    tab_width = 8
    indent_width = 4
    indent_tab = False
    auto_indent = True

    closed = False
    current_lastcommands = None
    theme = None
    highlight = None
    _check_fileupdate = 0
    
    def __init__(self):
        self.commands = {}
        self.is_availables = {}

        self.keybind = keyboard.KeyBind()
        self.keybind_vi_commandmode = keyboard.KeyBind()
        self.keybind_vi_visualmode = keyboard.KeyBind()
        self.keybind_vi_visuallinewisemode = keyboard.KeyBind()
        self.init_keybind()

        self.init_commands()
        self.init_menu()

        self.themes = []
        self.init_themes()
        self._build_theme()
        kaa.app.translate_theme(self.theme)
        
        self.tokenizers = []
        self.init_tokenizers()
        self.stylemap = {}
        self.highlight = highlight.Highlighter(self.tokenizers)

        self.setup()

    def setup(self):
        pass
        
    def close(self):
        self.document = None
        self.closed = True

        self.keybind.clear()

        self.theme = None

        self.commands.clear()
        self.commands = None

        self.is_availables = None
        self.highlight.close()
        self.highlight = None
        self.tokenizers = None
        self.stylemap = None

    def _build_style_map(self):
        self.stylemap[0] = self.theme.get_style('default')
        for tokenid in self.highlight.tokenids:
            assert tokenid not in self.stylemap
            token = self.highlight.get_token(tokenid)
            if token:
                stylename = token.get_stylename(tokenid)
                style = self.theme.get_style(stylename)

                self.stylemap[tokenid] = style
            else:
                self.stylemap[tokenid] = self.theme.get_style('default')

    def on_set_document(self, document):
        self.document = document
        self.document.styles.setints(0, len(self.document.styles), 0)

        self._build_style_map()
        self.document.use_undo(self.USE_UNDO)
        
    def on_document_updated(self, pos, inslen, dellen):
        if self.highlight:
            self.highlight.updated(self.document, pos, inslen, dellen)

    def on_file_saved(self, fileinfo):
        pass

    def on_add_window(self, wnd):
        self.editmode_insert(wnd)

    def on_del_window(self, wnd):
        if len(self.document.wnds) == 1:
            if self.DOCUMENT_MODE and self.document.fileinfo:
                self.document.fileinfo.save_screenloc(wnd.screen.pos)

    def on_focus(self, wnd):
        kaa.app.show_cursor(self.is_cursor_visible())

        # relocate cursor
        wnd.cursor.setpos(wnd.cursor.pos)

    def register_keys(self, keybind, keys):
        for d in keys:
            keybind.add_keybind(d)

    def init_keybind(self):
        pass

    def init_commands(self):
        for name in dir(self):
            attr = getattr(self, name)
            if hasattr(attr, 'COMMAND_ID') and callable(attr):
                self.commands[getattr(attr, 'COMMAND_ID')] = attr

    def init_menu(self):
        pass
        
    def init_themes(self):
        pass

    def init_tokenizers(self):
        pass

    def register_command(self, cmds):
        self.commands.update(cmds.get_commands())
        self.is_availables.update(cmds.get_commands_is_enable())

    def get_command(self, commandid):
        is_available = self.is_availables.get(commandid, None)
        cmd = self.commands.get(commandid, None)
        return (is_available, cmd)

    def editmode_insert(self, wnd):
        wnd.set_editmode(editmode.EditMode())

    def editmode_visual(self, wnd):
        wnd.set_editmode(editmode.VisualMode())

    def editmode_visual_linewise(self, wnd):
        wnd.set_editmode(editmode.VisualLinewiseMode())

    def editmode_command(self, wnd):
        wnd.set_editmode(editmode.CommandMode())

    def get_styleid(self, stylename):
        if stylename == 'default':
            return 0

        for styleid, style in self.stylemap.items():
            if style.name == stylename:
                return styleid
        else:
            if not self.stylemap:
                ret = 1
            else:
                ret = max(self.stylemap.keys())+1
            self.stylemap[ret] = self.theme.get_style(stylename)
            return ret

    def select_theme(self, theme_name, themes):
        theme = themes.get(theme_name, None)
        if theme is None:
            theme = themes[kaa.app.DEFAULT_THEME]
        return theme

    def _build_theme(self):
        theme_name = kaa.app.get_current_theme()
        self.theme = theme.Theme([])
        for t in self.themes:
            self.theme.update(self.select_theme(theme_name, t))

    def get_style(self, tokenid):
        ret = self.stylemap.get(tokenid, None)
        if ret is None:
            ret = self.theme.get_style('default')
        return ret

    def get_menu(self, itemname):
        return self.menu.get(itemname, None)
        
    def is_cursor_visible(self):
        return self.SHOW_CURSOR

    def on_keypressed(self, wnd, event, s, commands, candidate):
        return s, commands, candidate

    def filter_string(self, wnd, s):
        return s

    def on_edited(self, wnd):
        pass

    def on_str(self, wnd, s):
        self.edit_commands.put_string(wnd, s)
        wnd.screen.selection.clear()

        if kaa.app.macro.is_recording():
            kaa.app.macro.record_string(s)
        if self.highlight:
            # run highlighter a bit to display changes immediately.
            self.highlight.update_style(self.document, batch=50)

    def add_lastcommands(self, commandids):
        if self.current_lastcommands is not None:
            self.current_lastcommands.extend(commandids)

    def on_commands(self, wnd, commandids):
        try:
            if callable(commandids):
                commandids(wnd)
                return

            self.current_lastcommands = []
            for commandid in commandids:
                is_available, command = self.get_command(commandid)
                if not command:
                    msg = 'command {!r} is not registered.'.format(commandid)
                    kaa.app.messagebar.set_message(msg)
                    kaa.log.error(msg)
                    return

                command(wnd)
                if kaa.app.macro.is_recording():
                    kaa.app.macro.record(command)

                if not getattr(command, 'NORERUN', False):
                    self.current_lastcommands.append(commandid)

            if self.current_lastcommands:
                kaa.app.lastcommands = self.current_lastcommands

        finally:
            wnd.editmode.clear_repeat()
            self.current_lastcommands = None

    def on_esc_pressed(self, wnd, event):
        pass

    def on_cursor_located(self, wnd, pos, y, x):
        pass

    def update_charattr(self, wnd):
        if wnd.charattrs:
            wnd.charattrs.clear()
            wnd.screen.style_updated()
            return True

    def on_idle(self):
        if self.closed:
            return

        ret = self.run_highlight()
        return ret

    def get_line_overlays(self):
        return {}

    HIGHLIGHTBATCH = 300
    def run_highlight(self):
        if self.highlight:
            return self.highlight.update_style(self.document, batch=self.HIGHLIGHTBATCH)

    def _split_chars(self, begin, end):
        """split characters by character category."""

        s = self.document.gettext(begin, end)
        for key, chars in itertools.groupby(s, unicodedata.category):
            chars = ''.join(chars)
            end = begin + len(chars)
            yield begin, end, chars, key
            begin = end

    RE_WORDCHAR = r"(?P<L_WORDCHAR>[a-zA-Z0-9_]+)"
    RE_WHITESPACE = r"(?P<Z_WHITESPACE>[\t ]+)"
    RE_LF = r"(?P<_LF>\n)"
    RE_HIRAGANA = r"(?P<L_HIRAGANA>[\u3040-\u309f\u30fc]+)"
    RE_KATAKANA = r"(?P<L_KATAKANA>[\u30a0-\u30ff\u30fc]+)"

    RE_SPLITWORD = gappedbuf.re.compile('|'.join((
        RE_WORDCHAR, RE_WHITESPACE, RE_LF, RE_HIRAGANA, RE_KATAKANA)))

    def split_word(self, begin, all=False):
        """yield word in the document until line ends"""

        for m in self.RE_SPLITWORD.finditer(self.document.buf, pos=begin):
            # split unmatched characters by character category.
            f, t = m.span()
            if f != begin:
                yield from self._split_chars(begin, f)
            begin = t

            yield (f, t, m.group(), m.lastgroup)

            # finish if we reach '\n'
            if not all and m.lastgroup == '_LF':
                return

        yield from self._split_chars(begin, self.document.endpos())

    def get_word_at(self, pos):
        tol = self.document.gettol(pos)
        ret = None
        for f, t, s, cg in self.split_word(tol):
            if t <= pos:
                if cg != 'Z_WHITESPACE':
                    if t==pos and cg[0] == 'L':
                        return (f, t, cg)
                    else:
                        ret = (f, t, cg)
            elif cg == '_LF':
                break
            else:
                if cg == 'Z_WHITESPACE':
                    if f == pos:
                        return ret
                    else:
                        return
                return (f, t, cg)

        if ret and ret[1] == pos:
            return ret

    def get_word_list(self):
        words = collections.OrderedDict()
        for f, t, s, cg in self.split_word(0, all=True):
            if cg in {'Z_WHITESPACE', '_LF'}:
                continue
            if cg[0] in 'LMN': # Letter, Mark, Number
                if len(s) > 2:
                    words[s] = None
        return list(words.keys())

    def search_next(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        pos = min(max(0, pos), self.document.endpos())
        m = regex.search(self.document.buf, pos)
        return m

    def search_prev(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        last = None
        for m in regex.finditer(self.document.buf, 0):
            span = m.span()
            if span[1] >= pos:
                break
            last = m
        return last

    def get_line_sel(self, wnd):
        doc = wnd.document
        sel = wnd.screen.selection.get_selrange()
        if not sel:
            f = t = wnd.cursor.pos
        else:
            f, t = sel

        tol =  doc.gettol(f)

        t_tol =  doc.gettol(t)
        if t != t_tol:
            eol = doc.geteol(t)
        else:
            eol = t

        return tol, eol

    def get_indent_range(self, pos):
        tol =  self.document.gettol(pos)
        regex = gappedbuf.re.compile(self.RE_WHITESPACE)
        m = regex.match(self.document.buf, tol)
        if m:
            return m.span()
        else:
            return (tol, tol)

    def build_indent_str(self, col):
        if self.indent_tab:
            ctab = col // self.tab_width
            cspc = col % self.tab_width
            return  '\t' * ctab + ' ' * cspc
        else:
            return ' ' * col

    def on_auto_indent(self, wnd):
        pos = wnd.cursor.pos
        f, t = self.get_indent_range(pos)
        indent = self.document.gettext(f, min(pos, t))
        if pos <= t:
            self.edit_commands.insert_string(wnd, f, '\n', 
                    update_cursor=False)
            wnd.cursor.setpos(wnd.cursor.pos+1)
        else:
            indent = '\n'+indent
            self.edit_commands.insert_string(wnd, pos, indent, 
                    update_cursor=False)
            wnd.cursor.setpos(pos+len(indent))
        
        wnd.cursor.savecol()

    def calc_cols(self, f, t):
        chars = self.document.gettext(f, t)
        (dispchrs, dispcols, positions,
         intervals) = screen.translate_chars(f, chars, self.tab_width, 
                         kaa.app.config.AMBIGUOUS_WIDTH)

        return sum(dispcols)

    def on_global_next(self, wnd):
        return

    def on_global_prev(self, wnd):
        return
 


import os
from kaa import config

class TestHistory:
    filename = './test.db'
    def test_history(self):
        storage = config.KaaHistoryStorage(self.filename)
        try:
            hist = config.History('test', storage)
            storage.add_history(hist)

            hist.add('1')
            hist.add('2')
            hist.add('1')
            hist.add('2')

            assert hist.get() == [('2', None), ('1', None)]
            hist.close()
        finally:
            storage.close()
            os.unlink(self.filename)

    def test_histclose(self):
        storage = config.KaaHistoryStorage(self.filename)
        try:
            hist = config.History('test', storage)
            storage.add_history(hist)

            for i in range(config.History.MAX_HISTORY*2):
                hist.add(str(i))
            hist.close()
        finally:
            storage.close()

        storage = config.KaaHistoryStorage(self.filename)
        try:
            hist = config.History('test', storage)
            storage.add_history(hist)

            assert len(hist.get()) == config.History.MAX_HISTORY
            hist.close()
        finally:
            storage.close()
            os.unlink(self.filename)

           


from kaa.theme import Theme, Style, Overlay

DefaultThemes = {
    'basic':
        Theme([
            Style('default', 'default', 'default'),
            Style('lineno', 'White', 'Blue'),
            Style('parenthesis_cur', 'White', 'Blue'),
            Style('parenthesis_match', 'Red', 'Yellow'),

            Style('keyword', 'Magenta', 'default'),
            Style('constant', 'Red', 'default'),
            Style('directive', 'Orange', 'default'),
            Style('comment', 'Cyan', 'default'),
            Style('string', 'Blue', 'default'),
            Style('number', 'Green', 'default'),

            Overlay('cursor_row', None, 'Base02'),   
            Overlay('breakpoint', None, 'Base02'),   
            Overlay('current_row', None, 'Yellow'),   
        ]),
    
}



import socket
import pickle
import sys
import select
import curses
import kaa
import kaa.log
from . import keydef, color, dialog
from kaa import config, keyboard, document, macro
from kaa.ui.messagebar import messagebarmode

from kaa.exceptions import KaaError

class CuiApp:
    SHOW_MENU_MESSAGE = 'Type F1 or alt+/ for menu.'
    DEFAULT_THEME = 'basic'
    DEFAULT_PALETTE = 'dark'
    MAX_CLIPBOARD = 10
    
    def __init__(self, config):
        self.config = config
        self.colors = None
        self._idleprocs = None
        self.lastcommands = ()
        self.focus = None
        self._clipboard = []
        self._quit = False
        self.theme = self.DEFAULT_THEME
        self.last_dir = '.'
        self._input_readers = []
        
    def init(self, mainframe):
        if self.config.palette:
            self.set_palette(self.config.palette)
        elif not self.colors:
            self.set_palette(self.DEFAULT_PALETTE)
            
        self.config.init_history()
        
        self.messagebar = messagebarmode.MessageBarMode()

        buf = document.Buffer()
        doc = document.Document(buf)
        doc.setmode(self.messagebar)
        mainframe.set_messagebar(doc)

        self.mainframe = mainframe
        self.focus = self.mainframe
        self.macro = macro.Macro()

        self.mainframe.on_console_resized()
        self.messagebar.set_message(self.SHOW_MENU_MESSAGE)

    def on_shutdown(self):
        self.config.close()
        
    def get_current_theme(self):
        return self.theme

    def set_palette(self, name):
        palette = self.get_palette(name)
        self.colors = color.Colors(palette)

    def get_palette(self, name):
        if name == 'light':
            return color.LightPalette()
        else:
            return color.DarkPalette()
            
    def quit(self):
        self._quit = True

    def set_idlejob(self):
        self._idleprocs = [
            doc.mode.on_idle for doc in document.Document.all if doc.mode]

    def on_idle(self):
        if self._idleprocs:
            proc = self._idleprocs.pop(0)
            # proc() returns True if proc() still has job to be done.
            if proc():
                self._idleprocs.append(proc)
                return True

        if self._idleprocs:
            return True

    def translate_theme(self, theme):
        overlays = {}
        for name, overlay in theme.overlays.items():
            fg = bg = None
            if overlay.fgcolor:
                fg = self.colors.colornames.get(overlay.fgcolor.upper())
            if overlay.bgcolor:
                bg = self.colors.colornames.get(overlay.bgcolor.upper())
            overlays[name] = (fg, bg)

        for style in theme.styles.values():
            fg, bg = (self.colors.colornames.get(style.fgcolor.upper()),
                      self.colors.colornames.get(style.bgcolor.upper()))
            attr = self.colors.get_color(fg, bg)
            style.cui_colorattr = attr

            style.cui_overlays = {}
            for name, (o_fg, o_bg) in overlays.items():
                if o_fg is None:
                    o_fg = fg
                if o_bg is None:
                    o_bg = bg
                style.cui_overlays[name] = self.colors.get_color(o_fg, o_bg)

    def get_keyname(self, key):
        try:
            if not isinstance(key, int):
                key = ord(key)
                
            return str(curses.keyname(key), 'utf-8', 'replace')
        except Exception:
            return '?'

    def translate_key(self, mod, c):
        """Translate kaa's key value to curses keycode"""

        alt = keyboard.alt in mod
        ctrl = keyboard.ctrl in mod
        shift = keyboard.shift in mod

        if alt:
            meta = '\x1b'
        else:
            meta = ''

        if isinstance(c, str):
            if shift:
                raise KaaError(
                    'Cannot use shift key for character: {!r}'.format((mod, c)))
            if ctrl:
                c = c.upper()
                if not (0x40 <= ord(c) <= 0x5f):
                    raise KaaError(
                        'Cannot use control key for character: {!r}'.format((mod, c)))
                return meta+chr(ord(c)-0x40)
            else:
                return meta+c
        else:
            ret = keydef.keyfromname(c, ctrl, shift)
            if ret is None:
                raise KaaError(
                    'Cannot convert character: {!r}'.format((mod, c)))

            return [ret] if not meta else [meta, ret]

    def set_focus(self, wnd):
        if wnd is self.focus:
            return

        if self.focus:
            self.focus.on_killfocus()

        self.focus = wnd
        if wnd:
            wnd.on_focus()

    def show_doc(self, doc):
        '''
        Create new window for the doc and show it.
        '''
        ret = self.mainframe.show_doc(doc)
        return ret
        
    def show_inputline(self, doc):
        self._idleprocs = None  # Reschedule idle procs
        dlg = dialog.DialogWnd(parent=self.mainframe, doc=doc)
        self.mainframe.show_inputline(dlg)
        return dlg

    def show_dialog(self, doc):
        dlg = dialog.DialogWnd(parent=self.mainframe, doc=doc)
        self.mainframe.show_dialog(dlg)
        return dlg

    def get_frames(self):
        return self.mainframe.childframes[:]

    def get_activeframe(self):
        return self.mainframe.activeframe

    def show_cursor(self, f):
        try:
            curses.curs_set(f)
        except curses.error:
            # curses.curs_set() occasionally fails if $TERM=xterm-color
            pass

    def get_clipboard(self):
        return self._clipboard[0] if self._clipboard else ''
        
    def get_clipboards(self):
        return iter(self._clipboard)
        
    def set_clipboard(self, s):
        self._clipboard.insert(0, s)
        del self._clipboard[self.MAX_CLIPBOARD:]


    def add_input_reader(self, reader):
        self._input_readers.append(reader)

    def del_input_reader(self, reader):
        if reader in self._input_readers:
            self._input_readers.remove(reader)
        
    def run(self):
        
        nonblocking = True
        while not self._quit:
            try:
                if not self.focus:
                    kaa.log.error('Internal error: invalid focus window.')
                    break

                if not nonblocking:
                    # update screen before sleep.
                    curses.panel.update_panels()
                    curses.doupdate()

                rd = []
                for f in self._input_readers:
                    rd.extend(f.get_reader())

                try:
                    rlist, _, _ = select.select(
                            [sys.stdin]+rd, [], [], 
                            0 if nonblocking else None)

                except InterruptedError:
                    pass

                else:
                    ready = [r for r in rlist if r is not sys.stdin]
                    if ready:
                        nonblocking = True
                        for r in ready:
                            idx = rd.index(r)
                            self._input_readers[idx].read_input(r)

                inputs = self.focus.do_input(nonblocking=True)
                for c in inputs:
                    if isinstance(c, keydef.KeyEvent):
                        nonblocking = True
                        if c.key == curses.KEY_RESIZE:
                            self.mainframe.on_console_resized()
                            continue

                        if self.focus.editmode:
                            self.focus.editmode.on_keyevent(self.focus, c)

                if not inputs:
                    if self.mainframe.on_idle():
                        continue
                    # no input
                    if not self.on_idle():
                        if self.mainframe.on_idle():
                            continue
                        # No more idle jobs
                        nonblocking = False
                else:
                    self.set_idlejob()  # Reschedule idle procs
                    
            except Exception as e:
                kaa.log.error('Unhandled exception', exc_info=True)
                kaa.app.messagebar.set_message(str(e))

                nonblocking = True




import sys, os
KAA_VERSION = (0, 15, 1)

def version_info():
    return '''\
Kaa {major}.{minor}.{micro}
Copyright (c) 2013 Atsuo Ishimoto.
All Rights Reserved.

Python version: {python_version}
Platform: {platform}
$TERM: {term}
'''.format(
    major=KAA_VERSION[0], minor=KAA_VERSION[1],
    micro=KAA_VERSION[2],
    python_version=sys.version,
    platform=sys.platform,
    term=repr(os.environ.get('TERM', ''))[1:-1])



import copy, time
from collections import defaultdict, namedtuple
import kaa
from kaa.commands import (appcommand, filecommand, editorcommand, 
    editmodecommand, toolcommand)
from . import keybind, theme, modebase, menu
from kaa import highlight


class DefaultMode(modebase.ModeBase):
    DOCUMENT_MODE = True
    MODENAME = 'default'
    SHOW_LINENO = False
    KEY_BINDS = [
        keybind.app_keys,
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.addtional_edit_command_keys,
        keybind.emacs_keys,
        keybind.search_command_keys,
        keybind.macro_command_keys,
        keybind.rerun_keys,
    ]

    VI_KEY_BIND = [
#        keybind.command_mode_keys
    ]

    VI_VISUAL_MODE_KEY_BIND = [
#        keybind.visual_mode_keys
    ]
    VI_VISUAL_LINEWISE_MODE_KEY_BIND = [
#        keybind.visual_linewise_mode_keys
    ]

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.KEY_BINDS)
        self.register_keys(self.keybind_vi_commandmode, self.VI_KEY_BIND)
        self.register_keys(self.keybind_vi_visualmode,
                           self.VI_VISUAL_MODE_KEY_BIND)
        self.register_keys(self.keybind_vi_visuallinewisemode,
                           self.VI_VISUAL_LINEWISE_MODE_KEY_BIND)

    def init_commands(self):
        super().init_commands()

        self.app_commands = appcommand.ApplicationCommands()
        self.register_command(self.app_commands)

        self.file_commands = filecommand.FileCommands()
        self.register_command(self.file_commands)

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.code_commands = editorcommand.CodeCommands()
        self.register_command(self.code_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

        self.macro_commands = editorcommand.MacroCommands()
        self.register_command(self.macro_commands)

        self.search_commands = editorcommand.SearchCommands()
        self.register_command(self.search_commands)

        self.tools_commands = toolcommand.ToolCommands()
        self.register_command(self.tools_commands)

        self.mode_commands = editmodecommand.EditModeCommands()
        self.register_command(self.mode_commands)

    def init_menu(self):
        self.menu = copy.deepcopy(menu.MENUS)

    def init_themes(self):
        super().init_themes()
        self.themes.append(theme.DefaultThemes)

    def close(self):
        super().close()
        self.keybind_vi_commandmode.clear()
        self.keybind_vi_visualmode.clear()
        self.keybind_vi_visuallinewisemode.clear()


    def on_idle(self):
        if self.closed:
            return

        ret = super().on_idle()
        if not ret:
            ret = self.check_fileupdate()

        return ret

    INTERVAL_CHECKUPDATE = 15
    def check_fileupdate(self):
        if not self.DOCUMENT_MODE:
            return
        if not kaa.app.mainframe.is_idle():
            return

        t = time.time()
        if t - self._check_fileupdate < self.INTERVAL_CHECKUPDATE:
            return

        self._check_fileupdate = t
        if self.document.fileinfo:
            if self.document.fileinfo.check_update():
                self.file_commands.notify_fileupdated(self.document)
                
    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        return

        # Pressing esc key starts command mode.
        is_available, command = self.get_command('editmode.command')
        if command:
            command(wnd)
            if kaa.app.macro.is_recording():
                kaa.app.macro.record(command)

    def on_keypressed(self, wnd, event, s, commands, candidate):
        if not commands and not candidate:
            if not s or s[0] < ' ':
                kaa.app.messagebar.set_message(kaa.app.SHOW_MENU_MESSAGE)

        return super().on_keypressed(wnd, event, s, commands, candidate)

    def _show_parenthesis(self, charattrs, pos):
        charattrs[pos] = self.get_styleid('parenthesis_cur')
        matchpos = self.find_match_parenthesis(pos)
        if matchpos is not None:
            charattrs[matchpos] = self.get_styleid('parenthesis_match')
        
    def update_charattr(self, wnd):
        pos = wnd.cursor.pos
        d = {}
        c = ''
        if pos < self.document.endpos():
            c = self.document.buf[pos]

        if c and (c in self.PARENTHESIS):
            self._show_parenthesis(d, pos)
        elif 1 < pos:
            c = self.document.buf[pos-1]
            if c in self.PARENTHESIS_CLOSE:
                self._show_parenthesis(d, pos-1)
                    
        if d != wnd.charattrs:
            wnd.charattrs = d
            wnd.screen.style_updated()
            return True


    PARENTHESIS_OPEN = '({['
    PARENTHESIS_CLOSE = ')}]'
    PARENTHESIS = PARENTHESIS_OPEN + PARENTHESIS_CLOSE
    PARENSIS_PAIR = {o:c for (o, c) in
                     zip(PARENTHESIS_OPEN+PARENTHESIS_CLOSE,
                         PARENTHESIS_CLOSE+PARENTHESIS_OPEN)}

    def iter_parenthesis(self, posfrom):
        while True:
            pos = self.document.buf.findchr(
                self.PARENTHESIS, posfrom, self.document.endpos())

            if pos == -1:
                break

            attr = self.document.styles.getint(pos)
            yield pos, self.document.buf[pos], attr
            posfrom = pos+1

    def iter_rev_parenthesis(self, posfrom):
        posfrom += 1
        while True:
            pos = self.document.buf.rfindchr(
                self.PARENTHESIS, 0, posfrom)

            if pos == -1:
                break

            attr = self.document.styles.getint(pos)
            yield pos, self.document.buf[pos], attr
            posfrom = pos

    def find_match_parenthesis(self, posfrom):
        opener = self.document.buf[posfrom]
        curattr = self.document.styles.getint(posfrom)

        d = defaultdict(int)
        if opener in self.PARENTHESIS_OPEN:
            f = self.iter_parenthesis
            key = (opener, curattr)
        else:
            f = self.iter_rev_parenthesis
            key = (self.PARENSIS_PAIR[opener], curattr)


        for pos, c, attr in f(posfrom):
            if c in self.PARENTHESIS_OPEN:
                d[(c,attr)] += 1
            else:
                d[(self.PARENSIS_PAIR[c],attr)] -= 1

            if d.get(key) == 0:
                return pos

    _headerinfo = namedtuple('_headerinfo', 
            ['token', 'parent', 'name', 'dispname', 'lineno', 'pos'])
    class HeaderInfo(_headerinfo):
        def get_parents(self):
            ret = []
            p = self.parent
            while p:
                ret.insert(0, p)
                p = p.parent
            return ret

    def get_headers(self):
        return ()



import kaa
from kaa.command import Commands, command

class EditModeCommands(Commands):
    @command('editmode.command')
    def editmode_commandmode(self, wnd):
        wnd.document.mode.editmode_command(wnd)
        kaa.app.messagebar.set_message('You are in command mode now. Type `i` to return insert mode.')

    @command('editmode.insert')
    def editmode_insertmode(self, wnd):
        wnd.document.mode.editmode_insert(wnd)
        kaa.app.messagebar.set_message('You are in insert mode now. Type `ESC` to return command mode.')

    @command('editmode.visual')
    def editmode_visualmode(self, wnd):
        wnd.screen.selection.clear()
        wnd.document.mode.editmode_visual(wnd)

    @command('editmode.visual-linewise')
    def editmode_visualmode_linewise(self, wnd):
        wnd.screen.selection.clear()
        wnd.document.mode.screen_commands.select_cur_line(wnd)
        wnd.document.mode.editmode_visual_linewise(wnd)

